# Generated by Django 3.2.20 on 2023-07-27 21:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('edx_proctoring', '0024_delete_proctoredexamstudentattempthistory'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalproctoredexam',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical proctored exam', 'verbose_name_plural': 'historical proctored exams'},
        ),
        migrations.AlterModelOptions(
            name='historicalproctoredexamstudentattempt',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical proctored exam attempt', 'verbose_name_plural': 'historical proctored exam attempts'},
        ),
    ]
